"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    objectKindIntersection: ()=>objectKindIntersection,
    objectKindPathSegments: ()=>objectKindPathSegments,
    flattenObjectKind: ()=>flattenObjectKind
});
const _errorsJs = require("../../utils/errors.js");
const _composeJs = require("../compose.js");
const _nodeJs = require("../node.js");
const objectKindIntersection = (0, _composeJs.composeIntersection)((l, r, state)=>{
    if (typeof l === "string") {
        if (typeof r === "string") {
            return l === r ? (0, _composeJs.equality)() : state.addDisjoint("objectKind", l, r);
        }
        return l === r[0] ? r : state.addDisjoint("objectKind", l, r[0]);
    }
    if (typeof r === "string") {
        return l[0] === r ? l : state.addDisjoint("objectKind", l[0], r);
    }
    if (l[0] !== r[0]) {
        return state.addDisjoint("objectKind", l[0], r[0]);
    }
    const result = [
        l[0]
    ];
    let lImpliesR = true;
    let rImpliesL = true;
    for(let i = 1; i < l.length; i++){
        const lNode = l[i];
        const rNode = r[i];
        state.path.push(objectKindParameterToPathSegment(l[0], i));
        const parameterResult = (0, _nodeJs.nodeIntersection)(lNode, rNode, state);
        state.path.pop();
        if ((0, _composeJs.isEquality)(parameterResult)) {
            result[i] = lNode;
        } else if (parameterResult === l) {
            result[i] = lNode;
            rImpliesL = false;
        } else if (parameterResult === r) {
            result[i] = rNode;
            lImpliesR = false;
        } else if ((0, _composeJs.isDisjoint)(parameterResult)) {
            return parameterResult;
        } else {
            result[i] = parameterResult;
            lImpliesR = false;
            rImpliesL = false;
        }
    }
    return lImpliesR ? rImpliesL ? (0, _composeJs.equality)() : l : rImpliesL ? r : result;
});
const objectKindParameterToPathSegment = (objectKind, i)=>objectKind === "Array" ? "${number}" : objectKind === "Set" ? "${item}" : objectKind === "Map" ? i === 1 ? "${key}" : "${value}" : (0, _errorsJs.throwInternalError)(`Unexpected parameterized objectKind '${objectKind}'`);
const objectKindPathSegments = {
    "${number}": true,
    "${item}": true,
    "${key}": true,
    "${value}": true
};
const flattenObjectKind = (entries, rule, ctx)=>entries.push([
        "objectKind",
        typeof rule === "string" ? rule : rule[0] === "Map" ? [
            rule[0],
            (0, _nodeJs.flattenNode)(rule[1], ctx),
            (0, _nodeJs.flattenNode)(rule[2], ctx)
        ] : [
            rule[0],
            (0, _nodeJs.flattenNode)(rule[1], ctx)
        ]
    ]);
