"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    isOptional: ()=>isOptional,
    propsIntersection: ()=>propsIntersection,
    flattenProps: ()=>flattenProps
});
const _composeTs = require("../compose.js");
const _nodeTs = require("../node.js");
const isOptional = (prop)=>prop[0] === "?";
const nodeFrom = (prop)=>isOptional(prop) ? prop[1] : prop;
const propsIntersection = (0, _composeTs.composeIntersection)((0, _composeTs.composeKeyedIntersection)((propKey, l, r, context)=>{
    if (l === undefined) {
        return r === undefined ? (0, _composeTs.equality)() : r;
    }
    if (r === undefined) {
        return l;
    }
    context.path.push(propKey);
    const result = (0, _nodeTs.nodeIntersection)(nodeFrom(l), nodeFrom(r), context);
    context.path.pop();
    const resultIsOptional = isOptional(l) && isOptional(r);
    if ((0, _composeTs.isDisjoint)(result) && resultIsOptional) {
        // If an optional key has an empty intersection, the type can
        // still be satisfied as long as the key is not included. Set
        // the node to never rather than invalidating the type.
        return {};
    }
    return result;
}, {
    onEmpty: "bubble"
}));
const flattenProps = (entries, props, ctx)=>{
    const requiredProps = [];
    const optionalProps = [];
    for(const k in props){
        const prop = props[k];
        ctx.path.push(k);
        if (isOptional(prop)) {
            optionalProps.push([
                k,
                (0, _nodeTs.flattenNode)(prop[1], ctx)
            ]);
        } else {
            requiredProps.push([
                k,
                (0, _nodeTs.flattenNode)(prop, ctx)
            ]);
        }
        ctx.path.pop();
    }
    if (requiredProps.length) {
        entries.push([
            "requiredProps",
            requiredProps
        ]);
    }
    if (optionalProps.length) {
        entries.push([
            "optionalProps",
            optionalProps
        ]);
    }
};
