"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    branchIntersection: ()=>branchIntersection,
    rulesIntersection: ()=>rulesIntersection,
    narrowableRulesIntersection: ()=>narrowableRulesIntersection,
    precedenceMap: ()=>precedenceMap,
    flattenBranch: ()=>flattenBranch,
    literalSatisfiesRules: ()=>literalSatisfiesRules
});
const _astTs = require("../../parse/string/ast.js");
const _errorsTs = require("../../utils/errors.js");
const _genericsTs = require("../../utils/generics.js");
const _composeTs = require("../compose.js");
const _classTs = require("./class.js");
const _collapsibleSetTs = require("./collapsibleSet.js");
const _divisorTs = require("./divisor.js");
const _objectKindTs = require("./objectKind.js");
const _propsTs = require("./props.js");
const _rangeTs = require("./range.js");
const _regexTs = require("./regex.js");
const rulesOf = (branch)=>branch.input ?? branch;
const branchIntersection = (l, r, state)=>{
    const lRules = rulesOf(l);
    const rRules = rulesOf(r);
    const rulesResult = rulesIntersection(lRules, rRules, state);
    if ("morph" in l) {
        if ("morph" in r) {
            if (l.morph === r.morph) {
                return (0, _composeTs.isEquality)(rulesResult) || (0, _composeTs.isDisjoint)(rulesResult) ? rulesResult : {
                    input: rulesResult,
                    morph: l.morph
                };
            }
            return state.lastOperator === "&" ? (0, _errorsTs.throwParseError)((0, _astTs.writeImplicitNeverMessage)(state.path, "of morphs")) : {};
        }
        return (0, _composeTs.isDisjoint)(rulesResult) ? rulesResult : {
            input: (0, _composeTs.isEquality)(rulesResult) ? l.input : rulesResult,
            morph: l.morph
        };
    }
    if ("morph" in r) {
        return (0, _composeTs.isDisjoint)(rulesResult) ? rulesResult : {
            input: (0, _composeTs.isEquality)(rulesResult) ? r.input : rulesResult,
            morph: r.morph
        };
    }
    return rulesResult;
};
const rulesIntersection = (l, r, state)=>"value" in l ? "value" in r ? l.value === r.value ? (0, _composeTs.equality)() : state.addDisjoint("value", l.value, r.value) : literalSatisfiesRules(l.value, r, state) ? l : state.addDisjoint("leftAssignability", l, r) : "value" in r ? literalSatisfiesRules(r.value, l, state) ? r : state.addDisjoint("rightAssignability", l, r) : narrowableRulesIntersection(l, r, state);
const narrowIntersection = (0, _composeTs.composeIntersection)(_collapsibleSetTs.collapsibleListUnion);
const narrowableRulesIntersection = (0, _composeTs.composeKeyedIntersection)({
    objectKind: _objectKindTs.objectKindIntersection,
    divisor: _divisorTs.divisorIntersection,
    regex: _regexTs.regexIntersection,
    props: _propsTs.propsIntersection,
    class: _classTs.classIntersection,
    range: _rangeTs.rangeIntersection,
    narrow: narrowIntersection
}, {
    onEmpty: "bubble"
});
const ruleFlatteners = {
    objectKind: _objectKindTs.flattenObjectKind,
    regex: (entries, rule)=>{
        for (const source of (0, _genericsTs.listFrom)(rule)){
            entries.push([
                "regex",
                (0, _regexTs.getRegex)(source)
            ]);
        }
    },
    divisor: (entries, rule)=>{
        entries.push([
            "divisor",
            rule
        ]);
    },
    range: _rangeTs.flattenRange,
    class: (entries, rule)=>{
        entries.push([
            "class",
            rule
        ]);
    },
    props: _propsTs.flattenProps,
    narrow: (entries, rule)=>{
        for (const narrow of (0, _genericsTs.listFrom)(rule)){
            entries.push([
                "narrow",
                narrow
            ]);
        }
    },
    value: (entries, rule)=>{
        entries.push([
            "value",
            rule
        ]);
    }
};
const precedenceMap = {
    // Critical: No other checks are performed if these fail
    domain: 0,
    value: 0,
    domains: 0,
    branches: 0,
    objectKind: 0,
    switch: 0,
    alias: 0,
    // Shallow: All shallow checks will be performed even if one or more fail
    class: 1,
    regex: 1,
    divisor: 1,
    bound: 1,
    // Deep: Performed if all shallow checks pass, even if one or more deep checks fail
    requiredProps: 2,
    optionalProps: 2,
    // Narrow: Only performed if all shallow and deep checks pass
    narrow: 3,
    // Morph: Only performed if all validation passes
    morph: 4
};
const flattenBranch = (branch, ctx)=>{
    if ("morph" in branch) {
        const result = flattenRules(branch.input, ctx);
        result.push([
            "morph",
            branch.morph
        ]);
        return result;
    }
    return flattenRules(branch, ctx);
};
const flattenRules = (rules, ctx)=>{
    const entries = [];
    let k;
    for(k in rules){
        ruleFlatteners[k](entries, rules[k], ctx);
    }
    return entries.sort((l, r)=>precedenceMap[l[0]] - precedenceMap[r[0]]);
};
const literalSatisfiesRules = (data, rules, state)=>!state.type.meta.scope.type([
        "node",
        {
            [state.domain]: rules
        }
    ])(data).problems;
