"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    parseTuple: ()=>parseTuple,
    writeMalformedFunctionalExpressionMessage: ()=>writeMalformedFunctionalExpressionMessage
});
const _functorsTs = require("../../nodes/functors.js");
const _nodeTs = require("../../nodes/node.js");
const _domainsTs = require("../../utils/domains.js");
const _errorsTs = require("../../utils/errors.js");
const _definitionTs = require("../definition.js");
const _unenclosedTs = require("../string/shift/operand/unenclosed.js");
const _keyofTs = require("./keyof.js");
const _morphTs = require("./morph.js");
const _narrowTs = require("./narrow.js");
const parseTuple = (def, ctx)=>{
    if (isPostfixExpression(def)) {
        return postfixParsers[def[1]](def, ctx);
    }
    if (isPrefixExpression(def)) {
        return prefixParsers[def[0]](def, ctx);
    }
    const props = {};
    for(let i = 0; i < def.length; i++){
        ctx.path.push(`${i}`);
        props[i] = (0, _definitionTs.parseDefinition)(def[i], ctx);
        ctx.path.pop();
    }
    return {
        object: {
            objectKind: "Array",
            props,
            range: {
                comparator: "==",
                limit: def.length
            }
        }
    };
};
const parseBranchTuple = (def, ctx)=>{
    if (def[2] === undefined) {
        return (0, _errorsTs.throwParseError)((0, _unenclosedTs.writeMissingRightOperandMessage)(def[1], ""));
    }
    const l = (0, _definitionTs.parseDefinition)(def[0], ctx);
    const r = (0, _definitionTs.parseDefinition)(def[2], ctx);
    return def[1] === "&" ? (0, _nodeTs.intersection)(l, r, ctx.type) : (0, _nodeTs.union)(l, r, ctx.type);
};
const parseArrayTuple = (def, scope)=>_functorsTs.functors.Array((0, _definitionTs.parseDefinition)(def[0], scope));
const writeMalformedFunctionalExpressionMessage = (operator, rightDef)=>`Expression requires a function following '${operator}' (was ${typeof rightDef})`;
const isPostfixExpression = (def)=>postfixParsers[def[1]] !== undefined;
const postfixParsers = {
    "|": parseBranchTuple,
    "&": parseBranchTuple,
    "[]": parseArrayTuple,
    ":": _narrowTs.parseNarrowTuple,
    "=>": _morphTs.parseMorphTuple
};
const prefixParsers = {
    keyof: _keyofTs.parseKeyOfTuple,
    instanceof: (def)=>{
        if (typeof def[1] !== "function") {
            return (0, _errorsTs.throwParseError)(`Expected a constructor following 'instanceof' operator (was ${typeof def[1]}).`);
        }
        return {
            object: {
                class: def[1]
            }
        };
    },
    "===": (def)=>({
            [(0, _domainsTs.domainOf)(def[1])]: {
                value: def[1]
            }
        }),
    node: (def)=>def[1]
};
const isPrefixExpression = (def)=>prefixParsers[def[0]] !== undefined;
