"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    ArkTypeError: ()=>ArkTypeError,
    Problem: ()=>Problem,
    Problems: ()=>Problems,
    describeDomains: ()=>describeDomains,
    describeBranches: ()=>describeBranches,
    DataWrapper: ()=>DataWrapper,
    defaultProblemWriters: ()=>defaultProblemWriters,
    compileProblemOptions: ()=>compileProblemOptions
});
const _scannerTs = require("../parse/string/shift/scanner.js");
const _domainsTs = require("../utils/domains.js");
const _genericsTs = require("../utils/generics.js");
const _objectKindsTs = require("../utils/objectKinds.js");
const _pathsTs = require("../utils/paths.js");
const _serializeTs = require("../utils/serialize.js");
var _definitions_code, _definitions_code1, _definitions_code2, _clausesByPath, _pathKey;
class ArkTypeError extends TypeError {
    constructor(problems){
        super(problems.summary);
        this.cause = problems;
    }
}
class Problem {
    toString() {
        return this.message;
    }
    get message() {
        return this.writers.message(this.reason, this.path);
    }
    get reason() {
        return this.writers.reason(this.mustBe, this.was);
    }
    get mustBe() {
        return this.writers.mustBe(this.rule);
    }
    get was() {
        return this.writers.was(new DataWrapper(this.data));
    }
    constructor(code, path, data, rule, writers){
        this.code = code;
        this.path = path;
        this.data = data;
        this.rule = rule;
        this.writers = writers;
    }
}
class ProblemArray extends Array {
    add(code, data, input) {
        const problem = new Problem(code, // copy the path to avoid future mutations affecting it
        _pathsTs.Path.from(this.state.path), data, input, this.state.type.meta.problems[code]);
        const pathKey = `${this.state.path}`;
        // TODO: Fix multi
        // const existing = this.byPath[pathKey]
        // if (existing) {
        //     if (existing.code === "multi") {
        //         existing.message += `\n• ${problem.message}`
        //     } else {
        //         this.byPath[pathKey] = new Problem(
        //             "multi",
        //             existing.path,
        //             `• ${existing.message}\n• ${problem.message}`
        //         )
        //     }
        // } else {
        // }
        this.byPath[pathKey] = problem;
        this.push(problem);
    }
    get summary() {
        if (this.length === 1) {
            return this[0].message;
        }
        return this.join("\n");
    }
    toString() {
        return this.summary;
    }
    throw() {
        throw new ArkTypeError(this);
    }
    constructor(state){
        super();
        this.state = state;
        this.byPath = {};
    }
}
const Problems = ProblemArray;
const capitalize = (s)=>s[0].toUpperCase() + s.slice(1);
const describeDomains = (domains)=>{
    if (domains.length === 1) {
        return _domainsTs.domainDescriptions[domains[0]];
    }
    if (domains.length === 0) {
        return "never";
    }
    return describeBranches(domains.map((objectKind)=>_domainsTs.domainDescriptions[objectKind]));
};
const describeBranches = (descriptions)=>{
    let description = "";
    for(let i = 0; i < descriptions.length - 1; i++){
        description += descriptions[i];
        if (i < descriptions.length - 2) {
            description += ", ";
        }
    }
    description += ` or ${descriptions[descriptions.length - 1]}`;
    return description;
};
class DataWrapper {
    toString() {
        return (0, _serializeTs.stringify)(this.value);
    }
    get domain() {
        return (0, _domainsTs.domainOf)(this.value);
    }
    get objectKind() {
        return (0, _objectKindsTs.objectKindOf)(this.value) ?? "non-object";
    }
    get size() {
        return typeof this.value === "string" || Array.isArray(this.value) ? this.value.length : typeof this.value === "number" ? this.value : 0;
    }
    get units() {
        return typeof this.value === "string" ? "characters" : Array.isArray(this.value) ? "items" : "";
    }
    get className() {
        return Object(this.value).constructor.name;
    }
    constructor(value){
        this.value = value;
    }
}
const writeDefaultWasDescription = (data)=>`${data}`;
const writeDefaultProblemReason = (mustBe, was)=>`must be ${mustBe}${was && ` (was ${was})`}`;
const writeDefaultProblemMessage = (reason, path)=>path.length ? `${path} ${reason}` : capitalize(reason);
const compileDefaultProblemWriters = (definitions)=>{
    let code;
    for(code in definitions){
        (_definitions_code = definitions[code]).was ?? (_definitions_code.was = writeDefaultWasDescription);
        (_definitions_code1 = definitions[code]).reason ?? (_definitions_code1.reason = writeDefaultProblemReason);
        (_definitions_code2 = definitions[code]).message ?? (_definitions_code2.message = writeDefaultProblemMessage);
    }
    return definitions;
};
const defaultProblemWriters = compileDefaultProblemWriters({
    divisor: {
        mustBe: (divisor)=>divisor === 1 ? `an integer` : `divisible by ${divisor}`
    },
    class: {
        mustBe: (constructor)=>`an instance of ${constructor.name}`,
        was: (data)=>data.className
    },
    domain: {
        mustBe: (domain)=>_domainsTs.domainDescriptions[domain],
        was: (data)=>data.domain
    },
    domainBranches: {
        mustBe: (domains)=>describeDomains(domains),
        was: (data)=>data.domain
    },
    objectKind: {
        mustBe: (kind)=>_objectKindsTs.objectKindDescriptions[kind],
        was: (data)=>data.objectKind
    },
    missing: {
        mustBe: ()=>"defined",
        was: ()=>""
    },
    bound: {
        mustBe: (bound)=>`${_scannerTs.Scanner.comparatorDescriptions[bound.comparator]} ${bound.limit}${bound.units ? ` ${bound.units}` : ""}`,
        was: (data)=>`${data.size}`
    },
    regex: {
        mustBe: (regex)=>`a string matching /${regex.source}/`
    },
    value: {
        mustBe: _serializeTs.stringify
    },
    valueBranches: {
        mustBe: (values)=>describeBranches(values.map(_serializeTs.stringify))
    },
    branches: {
        mustBe: (branchProblems)=>{
            const clausesByPath = {};
            for (const problem of branchProblems){
                const pathKey = `${problem.path}`;
                (_clausesByPath = clausesByPath)[_pathKey = pathKey] ?? (_clausesByPath[_pathKey] = []);
                clausesByPath[pathKey].push(problem.mustBe);
            }
            return describeBranches(Object.entries(clausesByPath).map(([path, clauses])=>`${path} must be ${clauses.length === 1 ? clauses[0] : describeBranches(clauses)}`));
        },
        reason: (mustBe, data)=>`${mustBe} (was ${data})`,
        message: (reason, path)=>path.length ? `At ${path}, ${reason}` : reason
    },
    multi: {
        mustBe: (problems)=>"• " + problems.join("\n• "),
        reason: (mustBe)=>mustBe
    }
});
const codes = (0, _genericsTs.keysOf)(defaultProblemWriters);
const compileProblemOptions = (opts, base = defaultProblemWriters)=>{
    if (!opts) {
        return base;
    }
    const { all , ...byCode } = opts;
    const result = {};
    let code;
    if (all) {
        for (code of codes){
            result[code] = {
                ...base[code],
                ...all,
                ...byCode[code]
            };
        }
    } else {
        for (code of codes){
            result[code] = byCode[code] ? {
                ...base[code],
                ...byCode[code]
            } : base[code];
        }
    }
    return result;
};
