"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    TraversalState: ()=>TraversalState,
    traverse: ()=>traverse,
    checkEntries: ()=>checkEntries,
    checkObjectKind: ()=>checkObjectKind
});
const _discriminateJs = require("../nodes/discriminate.js");
const _classJs = require("../nodes/rules/class.js");
const _divisorJs = require("../nodes/rules/divisor.js");
const _rangeJs = require("../nodes/rules/range.js");
const _regexJs = require("../nodes/rules/regex.js");
const _rulesJs = require("../nodes/rules/rules.js");
const _domainsJs = require("../utils/domains.js");
const _errorsJs = require("../utils/errors.js");
const _genericsJs = require("../utils/generics.js");
const _objectKindsJs = require("../utils/objectKinds.js");
const _pathsJs = require("../utils/paths.js");
const _serializeJs = require("../utils/serialize.js");
const _problemsJs = require("./problems.js");
function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
var _seen = /*#__PURE__*/ new WeakMap();
class TraversalState {
    traverseResolution(data, name) {
        const resolution = this.type.meta.scope.resolve(name);
        const id = resolution.meta.id;
        const isObject = (0, _domainsJs.hasDomain)(data, "object");
        if (isObject) {
            if ((0, _genericsJs.hasKey)(_classPrivateFieldGet(this, _seen), id)) {
                if (_classPrivateFieldGet(this, _seen)[id].includes(data)) {
                    // if data has already been checked by this alias as part of
                    // a resolution higher up on the call stack, it must be valid
                    // or we wouldn't be here
                    return;
                }
                _classPrivateFieldGet(this, _seen)[id].push(data);
            } else {
                _classPrivateFieldGet(this, _seen)[id] = [
                    data
                ];
            }
        }
        // TODO: type not an accurate reflection of context because not always included
        const lastType = this.type;
        this.type = resolution;
        traverse(data, resolution.flat, this);
        this.type = lastType;
        if (isObject) {
            _classPrivateFieldGet(this, _seen)[id].pop();
        }
    }
    traverseBranches(data, branches) {
        const lastProblems = this.problems;
        const lastPath = this.path;
        const lastFailFast = this.failFast;
        this.failFast = true;
        const branchProblems = [];
        for (const branch of branches){
            this.problems = new _problemsJs.Problems(this);
            this.path = new _pathsJs.Path();
            checkEntries(data, branch, this);
            if (!this.problems.length) {
                break;
            }
            branchProblems.push(this.problems[0]);
        }
        this.failFast = lastFailFast;
        this.path = lastPath;
        this.problems = lastProblems;
        if (branchProblems.length === branches.length) {
            this.problems.add("branches", data, branchProblems);
        }
    }
    constructor(type){
        _classPrivateFieldInit(this, _seen, {
            writable: true,
            value: void 0
        });
        this.type = type;
        this.path = new _pathsJs.Path();
        this.problems = new _problemsJs.Problems(this);
        this.failFast = false;
        _classPrivateFieldSet(this, _seen, {});
    }
}
const traverse = (data, node, state)=>{
    if (typeof node === "string") {
        if ((0, _domainsJs.domainOf)(data) !== node) {
            return state.problems.add("domain", data, node);
        }
        return;
    }
    return checkEntries(data, node, state);
};
const checkEntries = (data, entries, state)=>{
    let problemsPrecedence = Number.POSITIVE_INFINITY;
    const initialProblemsCount = state.problems.length;
    for (const [k, v] of entries){
        if (_rulesJs.precedenceMap[k] > problemsPrecedence) {
            return;
        }
        if (k === "morph") {
            if (typeof v === "function") {
                return v(data);
            }
            let out = data;
            for (const morph of v){
                out = morph(out);
            }
            return out;
        } else {
            checkers[k](data, v, state);
            if (state.problems.length > initialProblemsCount) {
                if (state.failFast) {
                    return;
                }
                problemsPrecedence = _rulesJs.precedenceMap[k];
            }
        }
    }
    return data;
};
const createPropChecker = (propKind)=>(data, props, state)=>{
        const rootPath = state.path;
        for (const [propKey, propNode] of props){
            state.path.push(propKey);
            if (!(0, _genericsJs.hasKey)(data, propKey)) {
                if (propKind !== "optionalProps") {
                    state.problems.add("missing", undefined, undefined);
                }
            } else {
                traverse(data[propKey], propNode, state);
            }
            state.path.pop();
        }
        state.path = rootPath;
    };
const checkRequiredProps = createPropChecker("requiredProps");
const checkOptionalProps = createPropChecker("optionalProps");
const checkObjectKind = (data, rule, state)=>{
    const dataObjectKind = (0, _objectKindsJs.objectKindOf)(data);
    if (typeof rule === "string") {
        if (dataObjectKind !== rule) {
            return state.problems.add("objectKind", data, rule);
        }
        return;
    }
    if (dataObjectKind !== rule[0]) {
        return state.problems.add("objectKind", data, rule[0]);
    }
    if (dataObjectKind === "Array" && typeof rule[2] === "number") {
        const actual = data.length;
        const expected = rule[2];
        if (expected !== actual) {
            return state.problems.add("bound", data, {
                comparator: "==",
                limit: expected,
                units: "items"
            });
        }
    }
    if (dataObjectKind === "Array" || dataObjectKind === "Set") {
        let i = 0;
        for (const item of data){
            state.path.push(`${i}`);
            traverse(item, rule[1], state);
            state.path.pop();
            i++;
        }
    } else {
        return (0, _errorsJs.throwInternalError)(`Unexpected objectKind entry ${(0, _serializeJs.stringify)(rule)}`);
    }
    return true;
};
const checkers = {
    regex: _regexJs.checkRegex,
    divisor: _divisorJs.checkDivisor,
    domains: (data, domains, state)=>{
        const entries = domains[(0, _domainsJs.domainOf)(data)];
        if (entries) {
            checkEntries(data, entries, state);
        } else {
            state.problems.add("domainBranches", data, (0, _genericsJs.keysOf)(domains));
        }
    },
    domain: (data, domain, state)=>{
        if ((0, _domainsJs.domainOf)(data) !== domain) {
            state.problems.add("domain", data, domain);
        }
    },
    objectKind: checkObjectKind,
    bound: _rangeJs.checkBound,
    requiredProps: checkRequiredProps,
    optionalProps: checkOptionalProps,
    branches: (data, branches, state)=>state.traverseBranches(data, branches),
    switch: (data, rule, state)=>{
        const dataAtPath = (0, _pathsJs.getPath)(data, rule.path);
        const caseKey = (0, _discriminateJs.serializeCase)(rule.kind, dataAtPath);
        if ((0, _genericsJs.hasKey)(rule.cases, caseKey)) {
            return checkEntries(data, rule.cases[caseKey], state);
        }
        const caseKeys = (0, _genericsJs.keysOf)(rule.cases);
        const lastPath = state.path;
        state.path = state.path.concat(rule.path);
        if (rule.kind === "value") {
            state.problems.add("valueBranches", dataAtPath, caseKeys.map((k)=>(0, _serializeJs.deserializePrimitive)(k)));
        } else {
            state.problems.add("domainBranches", dataAtPath, caseKeys);
        }
        state.path = lastPath;
    },
    alias: (data, name, state)=>state.traverseResolution(data, name),
    class: _classJs.checkClass,
    narrow: (data, narrow, state)=>narrow(data, state.problems),
    value: (data, value, state)=>{
        if (data !== value) {
            state.problems.add("value", data, value);
        }
    }
};
