function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classApplyDescriptorUpdate(receiver, descriptor) {
    if (descriptor.set) {
        if (!descriptor.get) {
            throw new TypeError("attempted to read set only private field");
        }
        if (!("__destrWrapper" in descriptor)) {
            descriptor.__destrWrapper = {
                set value (v){
                    descriptor.set.call(receiver, v);
                },
                get value () {
                    return descriptor.get.call(receiver);
                }
            };
        }
        return descriptor.__destrWrapper;
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        return descriptor;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
function _classPrivateFieldUpdate(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "update");
    return _classApplyDescriptorUpdate(receiver, descriptor);
}
function _classPrivateMethodGet(receiver, privateSet, fn) {
    if (!privateSet.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return fn;
}
function _classPrivateMethodInit(obj, privateSet) {
    _checkPrivateRedeclaration(obj, privateSet);
    privateSet.add(obj);
}
var _ctx_type_meta;
import { flattenType } from "./nodes/node.js";
import { parseDefinition } from "./parse/definition.js";
import { compileProblemOptions } from "./traverse/problems.js";
import { TraversalState, traverse } from "./traverse/traverse.js";
import { chainableNoOpProxy } from "./utils/chainableNoOpProxy.js";
import { throwInternalError, throwParseError } from "./utils/errors.js";
import { deepFreeze } from "./utils/freeze.js";
import { Path } from "./utils/paths.js";
let anonymousScopeCount = 0;
const scopeRegistry = {};
var _resolutions = /*#__PURE__*/ new WeakMap(), _exports = /*#__PURE__*/ new WeakMap(), _anonymousTypeCount = /*#__PURE__*/ new WeakMap(), _register = /*#__PURE__*/ new WeakSet(), _cacheSpaces = /*#__PURE__*/ new WeakSet(), _initializeContext = /*#__PURE__*/ new WeakSet(), _compiled = /*#__PURE__*/ new WeakMap(), _resolveRecurse = /*#__PURE__*/ new WeakSet();
export class Scope {
    createAnonymousTypeName() {
        return `type${++_classPrivateFieldUpdate(this, _anonymousTypeCount).value}`;
    }
    get infer() {
        return chainableNoOpProxy;
    }
    compile() {
        if (!_classPrivateFieldGet(this, _compiled)) {
            for(const name in this.aliases){
                this.resolve(name);
            }
            _classPrivateFieldSet(this, _compiled, true);
        }
        return _classPrivateFieldGet(this, _exports).root;
    }
    addReferenceIfResolvable(name, ctx) {
        const resolution = _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, name, "undefined", []);
        if (!resolution) {
            return false;
        }
        (_ctx_type_meta = ctx.type.meta).includesMorph || (_ctx_type_meta.includesMorph = resolution.meta.includesMorph);
        return true;
    }
    resolve(name) {
        return _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, name, "throw", []);
    }
    resolveNode(node) {
        if (typeof node === "object") {
            return node;
        }
        return this.resolveNode(this.resolve(node).node);
    }
    constructor(aliases, opts){
        _classPrivateMethodInit(this, _register);
        _classPrivateMethodInit(this, _cacheSpaces);
        _classPrivateMethodInit(this, _initializeContext);
        _classPrivateMethodInit(this, _resolveRecurse);
        _classPrivateFieldInit(this, _resolutions, {
            writable: true,
            value: void 0
        });
        _classPrivateFieldInit(this, _exports, {
            writable: true,
            value: void 0
        });
        _classPrivateFieldInit(this, _anonymousTypeCount, {
            writable: true,
            value: void 0
        });
        _classPrivateFieldInit(this, _compiled, {
            writable: true,
            value: void 0
        });
        this.aliases = aliases;
        this.opts = opts;
        this.parseCache = new FreezingCache();
        _classPrivateFieldSet(this, _resolutions, new Cache());
        _classPrivateFieldSet(this, _exports, new Cache());
        _classPrivateFieldSet(this, _anonymousTypeCount, 0);
        this.type = (def, opts = {})=>{
            const result = initializeType(def, opts, this);
            const ctx = _classPrivateMethodGet(this, _initializeContext, initializeContext).call(this, result);
            result.node = this.resolveNode(parseDefinition(def, ctx));
            result.flat = flattenType(result);
            return result;
        };
        _classPrivateFieldSet(this, _compiled, false);
        this.name = _classPrivateMethodGet(this, _register, register).call(this, opts);
        if (opts.standard !== false) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, [
                standardTypes
            ], "imports");
        }
        if (opts.imports) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, opts.imports, "imports");
        }
        if (opts.includes) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, opts.includes, "includes");
        }
    }
}
function register(opts) {
    const name = opts.name ? scopeRegistry[opts.name] ? throwParseError(`A scope named '${opts.name}' already exists`) : opts.name : `anonymousScope${++anonymousScopeCount}`;
    scopeRegistry[name] = this;
    return name;
}
function cacheSpaces(spaces, kind) {
    for (const space of spaces){
        for(const name in space){
            if (_classPrivateFieldGet(this, _resolutions).has(name) || name in this.aliases) {
                throwParseError(writeDuplicateAliasMessage(name));
            }
            _classPrivateFieldGet(this, _resolutions).set(name, space[name]);
            if (kind === "includes") {
                _classPrivateFieldGet(this, _exports).set(name, space[name]);
            }
        }
    }
}
function initializeContext(type) {
    return {
        type,
        path: new Path()
    };
}
function resolveRecurse(name, onUnresolvable, seen) {
    const maybeCacheResult = _classPrivateFieldGet(this, _resolutions).get(name);
    if (maybeCacheResult) {
        return maybeCacheResult;
    }
    if (!this.aliases[name]) {
        return onUnresolvable === "throw" ? throwInternalError(`Unexpectedly failed to resolve alias '${name}'`) : undefined;
    }
    // TODO: opts?
    const type = initializeType(this.aliases[name], {
        name
    }, this);
    _classPrivateFieldGet(this, _resolutions).set(name, type);
    _classPrivateFieldGet(this, _exports).set(name, type);
    const ctx = _classPrivateMethodGet(this, _initializeContext, initializeContext).call(this, type);
    let resolution = parseDefinition(type.meta.definition, ctx);
    if (typeof resolution === "string") {
        if (seen.includes(resolution)) {
            return throwParseError(writeShallowCycleErrorMessage(name, seen));
        }
        seen.push(resolution);
        resolution = _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, resolution, "throw", seen).node;
    }
    type.node = resolution;
    type.flat = flattenType(type);
    return type;
}
// TODO: merge scope options
const initializeType = (definition, opts, scope)=>{
    const name = opts.name ?? "type";
    const meta = {
        name,
        id: `${scope.name}.${opts.name ? name : scope.createAnonymousTypeName()}`,
        definition,
        scope,
        problems: compileProblemOptions(opts.problems),
        includesMorph: false
    };
    const root = {
        // temporarily initialize node/flat to aliases that will be included in
        // the final type in case of cyclic resolutions
        node: name,
        flat: [
            [
                "alias",
                name
            ]
        ],
        meta,
        infer: chainableNoOpProxy
    };
    // dynamically assign a name to the primary traversal function
    const namedTraverse = {
        [name]: (data)=>{
            const state = new TraversalState(type);
            const out = traverse(data, type.flat, state);
            return state.problems.length ? {
                data,
                problems: state.problems
            } : {
                data,
                out
            };
        }
    }[name];
    // we need to assign this to a variable before returning so we can reference
    // it in namedTraverse
    const type = Object.assign(namedTraverse, root);
    return type;
};
class Cache {
    get root() {
        return this.cache;
    }
    has(name) {
        return name in this.cache;
    }
    get(name) {
        return this.cache[name];
    }
    set(name, item) {
        this.cache[name] = item;
        return item;
    }
    constructor(){
        this.cache = {};
    }
}
class FreezingCache extends Cache {
    set(name, item) {
        this.cache[name] = deepFreeze(item);
        return item;
    }
}
export const writeShallowCycleErrorMessage = (name, seen)=>`Alias '${name}' has a shallow resolution cycle: ${[
        ...seen,
        name
    ].join("=>")}`;
const always = {
    bigint: true,
    boolean: true,
    null: true,
    number: true,
    object: true,
    string: true,
    symbol: true,
    undefined: true
};
export const scope = (aliases, opts = {})=>new Scope(aliases, opts);
const tsKeywords = scope({
    any: [
        "node",
        always
    ],
    bigint: [
        "node",
        {
            bigint: true
        }
    ],
    boolean: [
        "node",
        {
            boolean: true
        }
    ],
    false: [
        "node",
        {
            boolean: {
                value: false
            }
        }
    ],
    never: [
        "node",
        {}
    ],
    null: [
        "node",
        {
            null: true
        }
    ],
    number: [
        "node",
        {
            number: true
        }
    ],
    object: [
        "node",
        {
            object: true
        }
    ],
    string: [
        "node",
        {
            string: true
        }
    ],
    symbol: [
        "node",
        {
            symbol: true
        }
    ],
    true: [
        "node",
        {
            boolean: {
                value: true
            }
        }
    ],
    unknown: [
        "node",
        always
    ],
    void: [
        "node",
        {
            undefined: true
        }
    ],
    undefined: [
        "node",
        {
            undefined: true
        }
    ]
}, {
    name: "ts",
    standard: false
});
const tsKeywordsSpace = tsKeywords.compile();
const jsObjects = scope({
    Function: [
        "node",
        {
            object: {
                objectKind: "Function"
            }
        }
    ],
    Array: [
        "node",
        {
            object: {
                objectKind: "Array"
            }
        }
    ],
    Date: [
        "node",
        {
            object: {
                objectKind: "Date"
            }
        }
    ],
    Error: [
        "node",
        {
            object: {
                objectKind: "Error"
            }
        }
    ],
    Map: [
        "node",
        {
            object: {
                objectKind: "Map"
            }
        }
    ],
    RegExp: [
        "node",
        {
            object: {
                objectKind: "RegExp"
            }
        }
    ],
    Set: [
        "node",
        {
            object: {
                objectKind: "Set"
            }
        }
    ],
    Object: [
        "node",
        {
            object: {
                objectKind: "Object"
            }
        }
    ],
    String: [
        "node",
        {
            object: {
                objectKind: "String"
            }
        }
    ],
    Number: [
        "node",
        {
            object: {
                objectKind: "Number"
            }
        }
    ],
    Boolean: [
        "node",
        {
            object: {
                objectKind: "Boolean"
            }
        }
    ],
    WeakMap: [
        "node",
        {
            object: {
                objectKind: "WeakMap"
            }
        }
    ],
    WeakSet: [
        "node",
        {
            object: {
                objectKind: "WeakSet"
            }
        }
    ],
    Promise: [
        "node",
        {
            object: {
                objectKind: "Promise"
            }
        }
    ]
}, {
    name: "jsObjects",
    standard: false
});
const jsObjectsSpace = jsObjects.compile();
const validation = scope({
    email: /^(.+)@(.+)\.(.+)$/,
    alphanumeric: /^[dA-Za-z]+$/,
    alpha: /^[A-Za-z]+$/,
    lowercase: /^[a-z]*$/,
    uppercase: /^[A-Z]*$/,
    integer: [
        "node",
        {
            number: {
                divisor: 1
            }
        }
    ]
}, {
    name: "validation",
    standard: false
});
const validationSpace = validation.compile();
const standard = scope({}, {
    name: "standard",
    includes: [
        tsKeywordsSpace,
        jsObjectsSpace,
        validationSpace
    ],
    standard: false
});
const standardTypes = standard.compile();
export const scopes = {
    tsKeywords,
    jsObjects,
    validation,
    standard
};
export const spaces = {
    tsKeywords: tsKeywordsSpace,
    jsObjects: jsObjectsSpace,
    validation: validationSpace,
    standard: standardTypes
};
export const type = scopes.standard.type;
export const writeDuplicateAliasMessage = (name)=>`Alias '${name}' is already defined`;
export const isType = (value)=>value?.infer === chainableNoOpProxy;
