import { throwInternalError } from "../../utils/errors.js";
import { composeIntersection, equality, isDisjoint, isEquality } from "../compose.js";
import { flattenNode, nodeIntersection } from "../node.js";
export const objectKindIntersection = composeIntersection((l, r, state)=>{
    if (typeof l === "string") {
        if (typeof r === "string") {
            return l === r ? equality() : state.addDisjoint("objectKind", l, r);
        }
        return l === r[0] ? r : state.addDisjoint("objectKind", l, r[0]);
    }
    if (typeof r === "string") {
        return l[0] === r ? l : state.addDisjoint("objectKind", l[0], r);
    }
    if (l[0] !== r[0]) {
        return state.addDisjoint("objectKind", l[0], r[0]);
    }
    const result = [
        l[0]
    ];
    let lImpliesR = true;
    let rImpliesL = true;
    for(let i = 1; i < l.length; i++){
        const lNode = l[i];
        const rNode = r[i];
        state.path.push(objectKindParameterToPathSegment(l[0], i));
        const parameterResult = nodeIntersection(lNode, rNode, state);
        state.path.pop();
        if (isEquality(parameterResult)) {
            result[i] = lNode;
        } else if (parameterResult === l) {
            result[i] = lNode;
            rImpliesL = false;
        } else if (parameterResult === r) {
            result[i] = rNode;
            lImpliesR = false;
        } else if (isDisjoint(parameterResult)) {
            return parameterResult;
        } else {
            result[i] = parameterResult;
            lImpliesR = false;
            rImpliesL = false;
        }
    }
    return lImpliesR ? rImpliesL ? equality() : l : rImpliesL ? r : result;
});
const objectKindParameterToPathSegment = (objectKind, i)=>objectKind === "Array" ? "${number}" : objectKind === "Set" ? "${item}" : objectKind === "Map" ? i === 1 ? "${key}" : "${value}" : throwInternalError(`Unexpected parameterized objectKind '${objectKind}'`);
export const objectKindPathSegments = {
    "${number}": true,
    "${item}": true,
    "${key}": true,
    "${value}": true
};
export const flattenObjectKind = (entries, rule, ctx)=>entries.push([
        "objectKind",
        typeof rule === "string" ? rule : rule[0] === "Map" ? [
            rule[0],
            flattenNode(rule[1], ctx),
            flattenNode(rule[2], ctx)
        ] : [
            rule[0],
            flattenNode(rule[1], ctx)
        ]
    ]);
