import { composeIntersection, composeKeyedIntersection, equality, isDisjoint } from "../compose.js";
import { flattenNode, nodeIntersection } from "../node.js";
export const isOptional = (prop)=>prop[0] === "?";
const nodeFrom = (prop)=>isOptional(prop) ? prop[1] : prop;
export const propsIntersection = composeIntersection(composeKeyedIntersection((propKey, l, r, context)=>{
    if (l === undefined) {
        return r === undefined ? equality() : r;
    }
    if (r === undefined) {
        return l;
    }
    context.path.push(propKey);
    const result = nodeIntersection(nodeFrom(l), nodeFrom(r), context);
    context.path.pop();
    const resultIsOptional = isOptional(l) && isOptional(r);
    if (isDisjoint(result) && resultIsOptional) {
        // If an optional key has an empty intersection, the type can
        // still be satisfied as long as the key is not included. Set
        // the node to never rather than invalidating the type.
        return {};
    }
    return result;
}, {
    onEmpty: "bubble"
}));
export const flattenProps = (entries, props, ctx)=>{
    const requiredProps = [];
    const optionalProps = [];
    for(const k in props){
        const prop = props[k];
        ctx.path.push(k);
        if (isOptional(prop)) {
            optionalProps.push([
                k,
                flattenNode(prop[1], ctx)
            ]);
        } else {
            requiredProps.push([
                k,
                flattenNode(prop, ctx)
            ]);
        }
        ctx.path.pop();
    }
    if (requiredProps.length) {
        entries.push([
            "requiredProps",
            requiredProps
        ]);
    }
    if (optionalProps.length) {
        entries.push([
            "optionalProps",
            optionalProps
        ]);
    }
};
