import { resolutionExtendsDomain } from "../../nodes/predicate.js";
import { throwParseError } from "../../utils/errors.js";
import { hasKey, listFrom } from "../../utils/generics.js";
import { parseDefinition } from "../definition.js";
export const parseKeyOfTuple = (def, ctx)=>{
    const resolution = ctx.type.meta.scope.resolveNode(parseDefinition(def[1], ctx));
    if (!resolutionExtendsDomain(resolution, "object")) {
        return throwParseError("never");
    }
    if (resolution.object === true) {
        return {
            string: true
        };
    }
    const branchKeys = [];
    const objectBranches = listFrom(resolution.object);
    const initialKeys = getPropsFromBranch(objectBranches[0]);
    for(let i = 1; i < objectBranches.length; i++){
        const keys = getPropsFromBranch(objectBranches[i]);
        if (!keys.length) {
            return throwParseError("never");
        }
        branchKeys.push(keys);
    }
    const result = [];
    initialKeys.forEach((key)=>{
        const hasKey = branchKeys.every((keySet)=>keySet.includes(key));
        if (hasKey) {
            result.push({
                value: key
            });
        }
    });
    return {
        string: result
    };
};
const getPropsFromBranch = (branch)=>hasKey(branch, "props") ? Object.keys(branch.props) : [];
