var _definitions_code, _definitions_code1, _definitions_code2, _clausesByPath, _pathKey;
import { Scanner } from "../parse/string/shift/scanner.js";
import { domainDescriptions, domainOf } from "../utils/domains.js";
import { keysOf } from "../utils/generics.js";
import { objectKindDescriptions, objectKindOf } from "../utils/objectKinds.js";
import { Path } from "../utils/paths.js";
import { stringify } from "../utils/serialize.js";
export class ArkTypeError extends TypeError {
    constructor(problems){
        super(problems.summary);
        this.cause = problems;
    }
}
export class Problem {
    toString() {
        return this.message;
    }
    get message() {
        return this.writers.message(this.reason, this.path);
    }
    get reason() {
        return this.writers.reason(this.mustBe, this.was);
    }
    get mustBe() {
        return this.writers.mustBe(this.rule);
    }
    get was() {
        return this.writers.was(new DataWrapper(this.data));
    }
    constructor(code, path, data, rule, writers){
        this.code = code;
        this.path = path;
        this.data = data;
        this.rule = rule;
        this.writers = writers;
    }
}
class ProblemArray extends Array {
    add(code, data, input) {
        const problem = new Problem(code, // copy the path to avoid future mutations affecting it
        Path.from(this.state.path), data, input, this.state.type.meta.problems[code]);
        const pathKey = `${this.state.path}`;
        // TODO: Fix multi
        // const existing = this.byPath[pathKey]
        // if (existing) {
        //     if (existing.code === "multi") {
        //         existing.message += `\n• ${problem.message}`
        //     } else {
        //         this.byPath[pathKey] = new Problem(
        //             "multi",
        //             existing.path,
        //             `• ${existing.message}\n• ${problem.message}`
        //         )
        //     }
        // } else {
        // }
        this.byPath[pathKey] = problem;
        this.push(problem);
    }
    get summary() {
        if (this.length === 1) {
            return this[0].message;
        }
        return this.join("\n");
    }
    toString() {
        return this.summary;
    }
    throw() {
        throw new ArkTypeError(this);
    }
    constructor(state){
        super();
        this.state = state;
        this.byPath = {};
    }
}
export const Problems = ProblemArray;
const capitalize = (s)=>s[0].toUpperCase() + s.slice(1);
export const describeDomains = (domains)=>{
    if (domains.length === 1) {
        return domainDescriptions[domains[0]];
    }
    if (domains.length === 0) {
        return "never";
    }
    return describeBranches(domains.map((objectKind)=>domainDescriptions[objectKind]));
};
export const describeBranches = (descriptions)=>{
    let description = "";
    for(let i = 0; i < descriptions.length - 1; i++){
        description += descriptions[i];
        if (i < descriptions.length - 2) {
            description += ", ";
        }
    }
    description += ` or ${descriptions[descriptions.length - 1]}`;
    return description;
};
export class DataWrapper {
    toString() {
        return stringify(this.value);
    }
    get domain() {
        return domainOf(this.value);
    }
    get objectKind() {
        return objectKindOf(this.value) ?? "non-object";
    }
    get size() {
        return typeof this.value === "string" || Array.isArray(this.value) ? this.value.length : typeof this.value === "number" ? this.value : 0;
    }
    get units() {
        return typeof this.value === "string" ? "characters" : Array.isArray(this.value) ? "items" : "";
    }
    get className() {
        return Object(this.value).constructor.name;
    }
    constructor(value){
        this.value = value;
    }
}
const writeDefaultWasDescription = (data)=>`${data}`;
const writeDefaultProblemReason = (mustBe, was)=>`must be ${mustBe}${was && ` (was ${was})`}`;
const writeDefaultProblemMessage = (reason, path)=>path.length ? `${path} ${reason}` : capitalize(reason);
const compileDefaultProblemWriters = (definitions)=>{
    let code;
    for(code in definitions){
        (_definitions_code = definitions[code]).was ?? (_definitions_code.was = writeDefaultWasDescription);
        (_definitions_code1 = definitions[code]).reason ?? (_definitions_code1.reason = writeDefaultProblemReason);
        (_definitions_code2 = definitions[code]).message ?? (_definitions_code2.message = writeDefaultProblemMessage);
    }
    return definitions;
};
export const defaultProblemWriters = compileDefaultProblemWriters({
    divisor: {
        mustBe: (divisor)=>divisor === 1 ? `an integer` : `divisible by ${divisor}`
    },
    class: {
        mustBe: (constructor)=>`an instance of ${constructor.name}`,
        was: (data)=>data.className
    },
    domain: {
        mustBe: (domain)=>domainDescriptions[domain],
        was: (data)=>data.domain
    },
    domainBranches: {
        mustBe: (domains)=>describeDomains(domains),
        was: (data)=>data.domain
    },
    objectKind: {
        mustBe: (kind)=>objectKindDescriptions[kind],
        was: (data)=>data.objectKind
    },
    missing: {
        mustBe: ()=>"defined",
        was: ()=>""
    },
    bound: {
        mustBe: (bound)=>`${Scanner.comparatorDescriptions[bound.comparator]} ${bound.limit}${bound.units ? ` ${bound.units}` : ""}`,
        was: (data)=>`${data.size}`
    },
    regex: {
        mustBe: (regex)=>`a string matching /${regex.source}/`
    },
    value: {
        mustBe: stringify
    },
    valueBranches: {
        mustBe: (values)=>describeBranches(values.map(stringify))
    },
    branches: {
        mustBe: (branchProblems)=>{
            const clausesByPath = {};
            for (const problem of branchProblems){
                const pathKey = `${problem.path}`;
                (_clausesByPath = clausesByPath)[_pathKey = pathKey] ?? (_clausesByPath[_pathKey] = []);
                clausesByPath[pathKey].push(problem.mustBe);
            }
            return describeBranches(Object.entries(clausesByPath).map(([path, clauses])=>`${path} must be ${clauses.length === 1 ? clauses[0] : describeBranches(clauses)}`));
        },
        reason: (mustBe, data)=>`${mustBe} (was ${data})`,
        message: (reason, path)=>path.length ? `At ${path}, ${reason}` : reason
    },
    multi: {
        mustBe: (problems)=>"• " + problems.join("\n• "),
        reason: (mustBe)=>mustBe
    }
});
const codes = keysOf(defaultProblemWriters);
// TODO: remove all copy from codes?
export const compileProblemOptions = (opts, base = defaultProblemWriters)=>{
    if (!opts) {
        return base;
    }
    const { all , ...byCode } = opts;
    const result = {};
    let code;
    if (all) {
        for (code of codes){
            result[code] = {
                ...base[code],
                ...all,
                ...byCode[code]
            };
        }
    } else {
        for (code of codes){
            result[code] = byCode[code] ? {
                ...base[code],
                ...byCode[code]
            } : base[code];
        }
    }
    return result;
};
