function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
import { serializeCase } from "../nodes/discriminate.js";
import { checkClass } from "../nodes/rules/class.js";
import { checkDivisor } from "../nodes/rules/divisor.js";
import { checkBound } from "../nodes/rules/range.js";
import { checkRegex } from "../nodes/rules/regex.js";
import { precedenceMap } from "../nodes/rules/rules.js";
import { domainOf, hasDomain } from "../utils/domains.js";
import { throwInternalError } from "../utils/errors.js";
import { hasKey, keysOf } from "../utils/generics.js";
import { objectKindOf } from "../utils/objectKinds.js";
import { getPath, Path } from "../utils/paths.js";
import { deserializePrimitive, stringify } from "../utils/serialize.js";
import { Problems } from "./problems.js";
var _seen = /*#__PURE__*/ new WeakMap();
export class TraversalState {
    traverseResolution(data, name) {
        const resolution = this.type.meta.scope.resolve(name);
        const id = resolution.meta.id;
        const isObject = hasDomain(data, "object");
        if (isObject) {
            if (hasKey(_classPrivateFieldGet(this, _seen), id)) {
                if (_classPrivateFieldGet(this, _seen)[id].includes(data)) {
                    // if data has already been checked by this alias as part of
                    // a resolution higher up on the call stack, it must be valid
                    // or we wouldn't be here
                    return;
                }
                _classPrivateFieldGet(this, _seen)[id].push(data);
            } else {
                _classPrivateFieldGet(this, _seen)[id] = [
                    data
                ];
            }
        }
        // TODO: type not an accurate reflection of context because not always included
        const lastType = this.type;
        this.type = resolution;
        traverse(data, resolution.flat, this);
        this.type = lastType;
        if (isObject) {
            _classPrivateFieldGet(this, _seen)[id].pop();
        }
    }
    traverseBranches(data, branches) {
        const lastProblems = this.problems;
        const lastPath = this.path;
        const lastFailFast = this.failFast;
        this.failFast = true;
        const branchProblems = [];
        for (const branch of branches){
            this.problems = new Problems(this);
            this.path = new Path();
            checkEntries(data, branch, this);
            if (!this.problems.length) {
                break;
            }
            branchProblems.push(this.problems[0]);
        }
        this.failFast = lastFailFast;
        this.path = lastPath;
        this.problems = lastProblems;
        if (branchProblems.length === branches.length) {
            this.problems.add("branches", data, branchProblems);
        }
    }
    constructor(type){
        _classPrivateFieldInit(this, _seen, {
            writable: true,
            value: void 0
        });
        this.type = type;
        this.path = new Path();
        this.problems = new Problems(this);
        this.failFast = false;
        _classPrivateFieldSet(this, _seen, {});
    }
}
export const traverse = (data, node, state)=>{
    if (typeof node === "string") {
        if (domainOf(data) !== node) {
            return state.problems.add("domain", data, node);
        }
        return;
    }
    return checkEntries(data, node, state);
};
export const checkEntries = (data, entries, state)=>{
    let problemsPrecedence = Number.POSITIVE_INFINITY;
    const initialProblemsCount = state.problems.length;
    for (const [k, v] of entries){
        if (precedenceMap[k] > problemsPrecedence) {
            return;
        }
        if (k === "morph") {
            if (typeof v === "function") {
                return v(data);
            }
            let out = data;
            for (const morph of v){
                out = morph(out);
            }
            return out;
        } else {
            checkers[k](data, v, state);
            if (state.problems.length > initialProblemsCount) {
                if (state.failFast) {
                    return;
                }
                problemsPrecedence = precedenceMap[k];
            }
        }
    }
    return data;
};
const createPropChecker = (propKind)=>(data, props, state)=>{
        const rootPath = state.path;
        for (const [propKey, propNode] of props){
            state.path.push(propKey);
            if (!hasKey(data, propKey)) {
                if (propKind !== "optionalProps") {
                    state.problems.add("missing", undefined, undefined);
                }
            } else {
                traverse(data[propKey], propNode, state);
            }
            state.path.pop();
        }
        state.path = rootPath;
    };
const checkRequiredProps = createPropChecker("requiredProps");
const checkOptionalProps = createPropChecker("optionalProps");
export const checkObjectKind = (data, rule, state)=>{
    const dataObjectKind = objectKindOf(data);
    if (typeof rule === "string") {
        if (dataObjectKind !== rule) {
            return state.problems.add("objectKind", data, rule);
        }
        return;
    }
    if (dataObjectKind !== rule[0]) {
        return state.problems.add("objectKind", data, rule[0]);
    }
    if (dataObjectKind === "Array" && typeof rule[2] === "number") {
        const actual = data.length;
        const expected = rule[2];
        if (expected !== actual) {
            return state.problems.add("bound", data, {
                comparator: "==",
                limit: expected,
                units: "items"
            });
        }
    }
    if (dataObjectKind === "Array" || dataObjectKind === "Set") {
        let i = 0;
        for (const item of data){
            state.path.push(`${i}`);
            traverse(item, rule[1], state);
            state.path.pop();
            i++;
        }
    } else {
        return throwInternalError(`Unexpected objectKind entry ${stringify(rule)}`);
    }
    return true;
};
const checkers = {
    regex: checkRegex,
    divisor: checkDivisor,
    domains: (data, domains, state)=>{
        const entries = domains[domainOf(data)];
        if (entries) {
            checkEntries(data, entries, state);
        } else {
            state.problems.add("domainBranches", data, keysOf(domains));
        }
    },
    domain: (data, domain, state)=>{
        if (domainOf(data) !== domain) {
            state.problems.add("domain", data, domain);
        }
    },
    objectKind: checkObjectKind,
    bound: checkBound,
    requiredProps: checkRequiredProps,
    optionalProps: checkOptionalProps,
    branches: (data, branches, state)=>state.traverseBranches(data, branches),
    switch: (data, rule, state)=>{
        const dataAtPath = getPath(data, rule.path);
        const caseKey = serializeCase(rule.kind, dataAtPath);
        if (hasKey(rule.cases, caseKey)) {
            return checkEntries(data, rule.cases[caseKey], state);
        }
        const caseKeys = keysOf(rule.cases);
        const lastPath = state.path;
        state.path = state.path.concat(rule.path);
        if (rule.kind === "value") {
            state.problems.add("valueBranches", dataAtPath, caseKeys.map((k)=>deserializePrimitive(k)));
        } else {
            state.problems.add("domainBranches", dataAtPath, caseKeys);
        }
        state.path = lastPath;
    },
    alias: (data, name, state)=>state.traverseResolution(data, name),
    class: checkClass,
    narrow: (data, narrow, state)=>narrow(data, state.problems),
    value: (data, value, state)=>{
        if (data !== value) {
            state.problems.add("value", data, value);
        }
    }
};
