import type { Type } from "../main.js";
import type { ParseContext } from "../parse/definition.js";
import type { Domain } from "../utils/domains.js";
import type { Dict, stringKeyOf } from "../utils/generics.js";
import type { Intersector } from "./compose.js";
import type { DiscriminatedSwitch } from "./discriminate.js";
import type { Predicate } from "./predicate.js";
import type { LiteralRules, MorphEntry, RuleEntry } from "./rules/rules.js";
export type TypeNode<$ = Dict> = Identifier<$> | ResolvedNode<$>;
export type Identifier<$ = Dict> = stringKeyOf<$>;
/** If scope is provided, we also narrow each predicate to match its domain.
 * Otherwise, we use a base predicate for all types, which is easier to
 * manipulate.*/
export type ResolvedNode<$ = Dict> = {
    readonly [domain in Domain]?: Predicate<domain, $>;
};
export declare const nodeIntersection: Intersector<TypeNode>;
export declare const intersection: (l: TypeNode, r: TypeNode, type: Type) => TypeNode;
export declare const union: (l: TypeNode, r: TypeNode, type: Type) => ResolvedNode;
export type TraversalNode = Domain | TraversalEntry[];
export type TraversalEntry = RuleEntry | MorphEntry | DomainsEntry | CyclicReferenceEntry | DomainEntry | BranchesEntry | SwitchEntry;
export type TraversalKey = TraversalEntry[0];
export type TraversalRule<k extends TraversalKey> = Extract<TraversalEntry, [
    k,
    unknown
]>[1];
export type CyclicReferenceEntry = ["alias", string];
export type DomainEntry = ["domain", Domain];
export type DomainsEntry = [
    "domains",
    {
        readonly [domain in Domain]?: TraversalEntry[];
    }
];
export type BranchesEntry = ["branches", TraversalEntry[][]];
export type SwitchEntry = ["switch", DiscriminatedSwitch];
export type FlattenContext = ParseContext & {
    lastDomain: Domain;
};
export declare const flattenType: (type: Type) => TraversalNode;
export declare const flattenNode: (node: TypeNode, ctx: FlattenContext) => TraversalNode;
export declare const isLiteralNode: <domain extends keyof {
    bigint: bigint;
    boolean: boolean;
    number: number;
    object: object;
    string: string;
    symbol: symbol;
    undefined: undefined;
    null: null;
}>(node: ResolvedNode, domain: domain) => node is { [_ in domain]: LiteralRules<domain>; };
