import type { Type } from "../main.js";
import type { Morph } from "../parse/tuple/morph.js";
import type { Domain } from "../utils/domains.js";
import type { CollapsibleList, defined, Dict, xor } from "../utils/generics.js";
import type { BranchesComparison } from "./branches.js";
import type { IntersectionResult, KeyIntersectionFn } from "./compose.js";
import { IntersectionState } from "./compose.js";
import type { FlattenContext, ResolvedNode, TraversalEntry } from "./node.js";
import type { LiteralRules, Rules } from "./rules/rules.js";
export type Predicate<domain extends Domain = Domain, $ = Dict> = string extends keyof $ ? true | CollapsibleList<Branch> : true | CollapsibleList<Branch<domain, $>>;
export type Branch<domain extends Domain = Domain, $ = Dict> = xor<Rules<domain, $>, {
    input: Rules<domain, $>;
    morph: CollapsibleList<Morph>;
}>;
export type PredicateComparison = IntersectionResult<Predicate> | BranchesComparison;
export declare const comparePredicates: (l: Predicate, r: Predicate, context: IntersectionState) => PredicateComparison;
export declare const predicateIntersection: KeyIntersectionFn<Required<ResolvedNode>>;
export declare const predicateUnion: (domain: Domain, l: Predicate, r: Predicate, type: Type) => true | {
    readonly objectKind?: import("./rules/objectKind.js").ObjectKindRule<Dict>;
    readonly regex?: CollapsibleList<string>;
    readonly divisor?: number;
    readonly range?: import("./rules/range.js").Range;
    readonly props?: import("./rules/props.js").PropsRule<Dict>;
    readonly class?: import("../utils/generics.js").constructor;
    readonly narrow?: import("./rules/rules.js").NarrowRule;
    input?: undefined;
    morph?: undefined;
} | {
    readonly value: unknown;
    input?: undefined;
    morph?: undefined;
} | {
    input: import("./rules/rules.js").NarrowableRules<Dict> | LiteralRules<keyof {
        bigint: bigint;
        boolean: boolean;
        number: number;
        object: object;
        string: string;
        symbol: symbol;
        undefined: undefined;
        null: null;
    }>;
    morph: CollapsibleList<Morph>;
    readonly objectKind?: undefined;
    readonly regex?: undefined;
    readonly divisor?: undefined;
    readonly range?: undefined;
    readonly props?: undefined;
    readonly class?: undefined;
    readonly narrow?: undefined;
} | {
    input: import("./rules/rules.js").NarrowableRules<Dict> | LiteralRules<keyof {
        bigint: bigint;
        boolean: boolean;
        number: number;
        object: object;
        string: string;
        symbol: symbol;
        undefined: undefined;
        null: null;
    }>;
    morph: CollapsibleList<Morph>;
    readonly value?: undefined;
} | readonly Branch<keyof {
    bigint: bigint;
    boolean: boolean;
    number: number;
    object: object;
    string: string;
    symbol: symbol;
    undefined: undefined;
    null: null;
}, Dict>[];
export declare const flattenPredicate: (predicate: Predicate, context: FlattenContext) => TraversalEntry[];
export declare const isLiteralCondition: (predicate: Predicate) => predicate is LiteralRules;
export type DomainSubtypeResolution<domain extends Domain> = {
    readonly [k in domain]: defined<ResolvedNode[domain]>;
};
export declare const resolutionExtendsDomain: <domain extends keyof {
    bigint: bigint;
    boolean: boolean;
    number: number;
    object: object;
    string: string;
    symbol: symbol;
    undefined: undefined;
    null: null;
}>(resolution: ResolvedNode, domain: domain) => resolution is DomainSubtypeResolution<domain>;
