import type { Morph } from "../../parse/tuple/morph.js";
import type { Narrow } from "../../parse/tuple/narrow.js";
import type { Domain, inferDomain } from "../../utils/domains.js";
import type { CollapsibleList, constructor, Dict } from "../../utils/generics.js";
import type { IntersectionState, Intersector } from "../compose.js";
import type { FlattenContext, TraversalEntry } from "../node.js";
import type { ObjectKindRule, TraversalObjectKindRule } from "./objectKind.js";
import type { PropsRule, TraversalOptionalProps, TraversalRequiredProps } from "./props.js";
import type { FlatBound, Range } from "./range.js";
export type NarrowableRules<$ = Dict> = {
    readonly objectKind?: ObjectKindRule<$>;
    readonly regex?: CollapsibleList<string>;
    readonly divisor?: number;
    readonly range?: Range;
    readonly props?: PropsRule<$>;
    readonly class?: constructor;
    readonly narrow?: NarrowRule;
};
export type LiteralRules<domain extends Domain = Domain> = {
    readonly value: inferDomain<domain>;
};
export type NarrowRule = CollapsibleList<Narrow>;
export type Branch<domain extends Domain = Domain, $ = Dict> = Rules<domain, $> | MorphBranch<domain, $>;
export type MorphBranch<domain extends Domain = Domain, $ = Dict> = {
    input: Rules<domain, $>;
    morph: CollapsibleList<Morph>;
};
export type FlatBranch = FlatRules | FlatMorphedBranch;
export type FlatRules = RuleEntry[];
type FlatMorphedBranch = [...rules: FlatRules, morph: MorphEntry];
export type RuleEntry = ["objectKind", TraversalObjectKindRule] | ["regex", RegExp] | ["divisor", number] | ["bound", FlatBound] | ["class", constructor] | TraversalRequiredProps | TraversalOptionalProps | ["narrow", Narrow] | ["value", unknown];
export type MorphEntry = ["morph", CollapsibleList<Morph>];
export type Rules<domain extends Domain = Domain, $ = Dict> = Domain extends domain ? NarrowableRules | LiteralRules : domain extends "object" ? defineRuleSet<domain, "objectKind" | "props" | "range" | "narrow" | "class", $> : domain extends "string" ? defineRuleSet<domain, "regex" | "range" | "narrow", $> : domain extends "number" ? defineRuleSet<domain, "divisor" | "range" | "narrow", $> : defineRuleSet<domain, "narrow", $>;
type defineRuleSet<domain extends Domain, keys extends keyof NarrowableRules, $> = Pick<NarrowableRules<$>, keys> | LiteralRules<domain>;
export declare const branchIntersection: Intersector<Branch>;
export declare const rulesIntersection: Intersector<Rules>;
export declare const narrowableRulesIntersection: Intersector<NarrowableRules<Dict>>;
export type FlattenAndPushRule<t> = (entries: RuleEntry[], rule: t, ctx: FlattenContext) => void;
export declare const precedenceMap: {
    readonly [k in TraversalEntry[0]]: number;
};
export declare const flattenBranch: (branch: Branch, ctx: FlattenContext) => FlatBranch;
export declare const literalSatisfiesRules: (data: unknown, rules: NarrowableRules, state: IntersectionState) => boolean;
export {};
