import type { FlatBound } from "../nodes/rules/range.js";
import type { Domain } from "../utils/domains.js";
import { domainOf } from "../utils/domains.js";
import type { constructor, evaluate, evaluateObject, instanceOf } from "../utils/generics.js";
import type { DefaultObjectKind } from "../utils/objectKinds.js";
import { objectKindOf } from "../utils/objectKinds.js";
import { Path } from "../utils/paths.js";
import type { ConstrainedRuleTraversalData, TraversalState } from "./traverse.js";
export declare class ArkTypeError extends TypeError {
    cause: Problems;
    constructor(problems: Problems);
}
export declare class Problem {
    code: ProblemCode;
    path: Path;
    private data;
    private rule;
    private writers;
    constructor(code: ProblemCode, path: Path, data: ProblemDataInput, rule: ProblemRuleInput, writers: ProblemWriterConfig<any>);
    toString(): string;
    get message(): string;
    get reason(): string;
    get mustBe(): string;
    get was(): string;
}
declare class ProblemArray extends Array<Problem> {
    private state;
    byPath: Record<string, Problem>;
    constructor(state: TraversalState);
    add<code extends ProblemCode>(code: code, data: ProblemDataInput<code>, input: ProblemRuleInput<code>): void;
    get summary(): string;
    toString(): string;
    throw(): never;
}
export declare const Problems: new (state: TraversalState) => readonly Problem[] & {
    [k in Exclude<keyof ProblemArray, keyof unknown[]>]: ProblemArray[k];
};
export type Problems = instanceOf<typeof Problems>;
export declare const describeDomains: (domains: Domain[]) => string;
export declare const describeBranches: (descriptions: string[]) => string;
type ProblemRuleInputs = {
    divisor: number;
    class: constructor;
    domain: Domain;
    domainBranches: Domain[];
    objectKind: DefaultObjectKind;
    missing: undefined;
    bound: FlatBound;
    regex: RegExp;
    value: unknown;
    valueBranches: unknown[];
    multi: string[];
    branches: Problem[];
};
export type ProblemCode = evaluate<keyof ProblemRuleInputs>;
export type ProblemRuleInput<code extends ProblemCode = ProblemCode> = ProblemRuleInputs[code];
type ProblemDataInputs = {
    [code in ProblemCode]: code extends keyof ConstrainedRuleTraversalData ? ConstrainedRuleTraversalData[code] : unknown;
};
export type ProblemDataInput<code extends ProblemCode = ProblemCode> = ProblemDataInputs[code];
export type ProblemOptions<code extends ProblemCode> = {
    mustBe?: RuleWriter<code>;
    was?: DataWriter<code>;
    reason?: ReasonWriter;
    message?: MessageWriter;
};
export type ProblemWriterConfig<code extends ProblemCode> = Required<ProblemOptions<code>>;
export type RuleWriter<code extends ProblemCode> = (context: ProblemRuleInputs[code]) => string;
export type DataWriter<code extends ProblemCode> = (data: DataWrapper<code extends keyof ConstrainedRuleTraversalData ? ConstrainedRuleTraversalData[code] : unknown>) => string;
export type ReasonWriter = (rule: string, data: string) => string;
export type MessageWriter = (reason: string, path: Path) => string;
export declare class DataWrapper<value = unknown> {
    value: value;
    constructor(value: value);
    toString(): string;
    get domain(): domainOf<value>;
    get objectKind(): "non-object" | NonNullable<objectKindOf<value, {
        Array: ArrayConstructor;
        Date: DateConstructor;
        Error: ErrorConstructor;
        Function: FunctionConstructor;
        Map: MapConstructor;
        RegExp: RegExpConstructor;
        Set: SetConstructor;
        Object: ObjectConstructor;
        String: StringConstructor;
        Number: NumberConstructor;
        Boolean: BooleanConstructor;
        WeakMap: WeakMapConstructor;
        WeakSet: WeakSetConstructor;
        Promise: PromiseConstructor;
    }>>;
    get size(): number;
    get units(): "" | "characters" | "items";
    get className(): any;
}
export declare const defaultProblemWriters: ProblemsConfig;
export type ProblemsOptions = evaluateObject<{
    all?: ProblemOptions<ProblemCode>;
} & {
    [code in ProblemCode]?: ProblemOptions<code>;
}>;
export type ProblemsConfig = {
    [code in ProblemCode]: ProblemWriterConfig<code>;
};
export declare const compileProblemOptions: (opts: ProblemsOptions | undefined, base?: ProblemsConfig) => ProblemsConfig;
export {};
